//----------------------------------------------------------------------------
//
// copyright 2013 Scott Duplichan
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
//----------------------------------------------------------------------------
#include <stdio.h>
#include <stdint.h>
#include <stdarg.h>
#include <math.h>
#include <mpfr.h>

// functions.s
void  x87_80bit_sin   (void *arg_80bit, void *result_80bit);
void  x87_80bit_cos   (void *arg_80bit, void *result_80bit);
void  x87_80bit_tan   (void *arg_80bit, void *result_80bit);
void  x87_80bit_atan  (void *arg_80bit, void *result_80bit);
void  x87_80bit_f2xm1 (void *arg_80bit, void *result_80bit);

//----------------------------------------------------------------------------
//
// formatMessage - return formatted message in static buffer
//
char *formatMessage (char *format, ...)
   {
   #define DIMENSION(array) (sizeof (array) / sizeof (array [0]))
   va_list marker;
   static char buffer [8][200];
   static int cycle;
   char *position = buffer [cycle];
   cycle++;
   if (cycle == DIMENSION (buffer)) cycle = 0;

   va_start (marker, format);
   vsprintf (position, format, marker);
   va_end (marker);
   return position;
   }

//----------------------------------------------------------------------------
// hexDump - format floating point value as hex, with 16-bit groupings

static char *hexDump (void *data, int size)
    {
    int words = size / 2;
    char *position;
    char buffer [64];
    int index;
    uint16_t *array16;
    
    array16 = data;
    position = buffer;

    for (index = words - 1; index >= 0; index--)
        {
        if (index != words - 1) position += sprintf (position, " ");
        position += sprintf (position, "%04X", array16 [index]);
        }
    return formatMessage (buffer);
    }

//----------------------------------------------------------------------------
// hexWithLongDouble - format long double as hex followed by decimal

static char *hexWithLongDouble (long double value)
    {
    mpfr_t mpfr_value;
    char buffer [64];

    mpfr_init2 (mpfr_value, 64);
    mpfr_set_ld (mpfr_value, value, MPFR_RNDN);
    mpfr_snprintf (buffer, sizeof buffer, "%1.20RG", mpfr_value);
    mpfr_clear (mpfr_value);
    return formatMessage ("%s (decimal %s)", hexDump (&value, 10), buffer);
    }

//----------------------------------------------------------------------------
// mpfr_scale_2ui - wrapper to combine MPFR functions mpfr_mul_2ui and mpfr_div_2ui

static int mpfr_scale_2ui (mpfr_t rop, mpfr_t op1, long int op2, mpfr_rnd_t rnd)
    {
    if (op2 >= 0) return mpfr_mul_2ui (rop, op1, op2, rnd);
    else return mpfr_div_2ui (rop, op1, -op2, rnd);
    }

//----------------------------------------------------------------------------
// return ulp value of difference, using the Intel definition of ulp:
// error = abs ((f(x) - F(x)) / 2^(k-63))

static long double ulpDiff (mpfr_t mpfr_correct, long double computed)
    {
    mpfr_t mpfr_diff, mpfr_computed, mpfr_temp;
    long double result;
    long k;

    mpfr_init2 (mpfr_diff, 1024);
    mpfr_init2 (mpfr_computed, 1024);
    mpfr_init2 (mpfr_temp, 1024);

    mpfr_set_ld (mpfr_computed, computed, MPFR_RNDN);
    mpfr_sub (mpfr_diff, mpfr_correct, mpfr_computed, MPFR_RNDN);
    mpfr_abs (mpfr_diff, mpfr_diff, MPFR_RNDN);
    mpfr_get_d_2exp (&k, mpfr_correct, MPFR_RNDN);
    k--;
    mpfr_scale_2ui (mpfr_temp, mpfr_diff, 63 - k, MPFR_RNDN);
    result = mpfr_get_ld (mpfr_temp, MPFR_RNDN);

    mpfr_clear (mpfr_diff);
    mpfr_clear (mpfr_computed);
    mpfr_clear (mpfr_temp);
    return result;
    }

//----------------------------------------------------------------------------
// formatUlp - format long double as integer if big, or as floating point otherwise

static char *formatUlp (long double value)
    {
    if (value < 9999 || value > 0xFFFFFFFFFFFFFFFFull) return formatMessage ("%g", (double) value);
    return formatMessage ("%llu", (uint64_t)value);
    }

//----------------------------------------------------------------------------
// printResult - print result of an x87 fpu test case

static void printResult (long double arg, long double actual, long double fpu, long double ulp, char *comment)
    {
    printf ("%s\n", comment);
    printf ("argument   %s\n", hexWithLongDouble (arg));
    printf ("actual     %s\n", hexWithLongDouble (actual));
    printf ("x87 fpu    %s\n", hexWithLongDouble (fpu));
    printf ("error      %s ulp\n\n", formatUlp (ulp));
    }

//----------------------------------------------------------------------------
// sinTest - test x87 fpu fsin instruction

long double sinTest (long double arg, char *comment)
    {
    long double fpu, actual, ulp;
    mpfr_t mpfr_arg, mpfr_actual;

    mpfr_init2 (mpfr_arg, 1024);
    mpfr_init2 (mpfr_actual, 1024);
    mpfr_set_ld (mpfr_arg, arg, MPFR_RNDN);
    mpfr_sin (mpfr_actual, mpfr_arg, MPFR_RNDN);
    actual = mpfr_get_ld (mpfr_actual, MPFR_RNDN);
    x87_80bit_sin (&arg, &fpu);
    ulp = ulpDiff (mpfr_actual, fpu);
    mpfr_clear (mpfr_arg);
    mpfr_clear (mpfr_actual);
    if (comment) printResult (arg, actual, fpu, ulp, comment);
    return ulp;
    }

//----------------------------------------------------------------------------
// cosTest - test x87 fpu fcos instruction

long double cosTest (long double arg, char *comment)
    {
    long double fpu, actual, ulp;
    mpfr_t mpfr_arg, mpfr_actual;

    mpfr_init2 (mpfr_arg, 1024);
    mpfr_init2 (mpfr_actual, 1024);
    mpfr_set_ld (mpfr_arg, arg, MPFR_RNDN);
    mpfr_cos (mpfr_actual, mpfr_arg, MPFR_RNDN);
    actual = mpfr_get_ld (mpfr_actual, MPFR_RNDN);
    x87_80bit_cos (&arg, &fpu);
    ulp = ulpDiff (mpfr_actual, fpu);
    mpfr_clear (mpfr_arg);
    mpfr_clear (mpfr_actual);
    if (comment) printResult (arg, actual, fpu, ulp, comment);
    return ulp;
    }

//----------------------------------------------------------------------------
// tanTest - test x87 fpu fptan instruction

long double tanTest (long double arg, char *comment)
    {
    long double fpu, actual, ulp;
    mpfr_t mpfr_arg, mpfr_actual;

    mpfr_init2 (mpfr_arg, 1024);
    mpfr_init2 (mpfr_actual, 1024);
    mpfr_set_ld (mpfr_arg, arg, MPFR_RNDN);
    mpfr_tan (mpfr_actual, mpfr_arg, MPFR_RNDN);
    actual = mpfr_get_ld (mpfr_actual, MPFR_RNDN);
    x87_80bit_tan (&arg, &fpu);
    ulp = ulpDiff (mpfr_actual, fpu);
    mpfr_clear (mpfr_arg);
    mpfr_clear (mpfr_actual);
    if (comment) printResult (arg, actual, fpu, ulp, comment);
    return ulp;
    }

//----------------------------------------------------------------------------
// atanTest - test x87 fpu fpatan instruction

long double atanTest (long double arg, char *comment)
    {
    long double fpu, actual, ulp;
    mpfr_t mpfr_arg, mpfr_actual;

    mpfr_init2 (mpfr_arg, 1024);
    mpfr_init2 (mpfr_actual, 1024);
    mpfr_set_ld (mpfr_arg, arg, MPFR_RNDN);
    mpfr_atan (mpfr_actual, mpfr_arg, MPFR_RNDN);
    actual = mpfr_get_ld (mpfr_actual, MPFR_RNDN);
    x87_80bit_atan (&arg, &fpu);
    ulp = ulpDiff (mpfr_actual, fpu);
    mpfr_clear (mpfr_arg);
    mpfr_clear (mpfr_actual);
    if (comment) printResult (arg, actual, fpu, ulp, comment);
    return ulp;
    }

//----------------------------------------------------------------------------
// f2xm1Test - test x87 fpu f2xm1 instruction

long double f2xm1Test (long double arg, char *comment)
    {
    long double fpu, actual, ulp;
    mpfr_t mpfr_arg, mpfr_actual;

    mpfr_init2 (mpfr_arg, 1024);
    mpfr_init2 (mpfr_actual, 1024);
    mpfr_set_ld (mpfr_arg, arg, MPFR_RNDN);
    mpfr_exp2 (mpfr_actual, mpfr_arg, MPFR_RNDN);
    mpfr_sub_si (mpfr_actual, mpfr_actual, 1, MPFR_RNDN);
    actual = mpfr_get_ld (mpfr_actual, MPFR_RNDN);
    x87_80bit_f2xm1 (&arg, &fpu);
    ulp = ulpDiff (mpfr_actual, fpu);
    mpfr_clear (mpfr_arg);
    mpfr_clear (mpfr_actual);
    if (comment) printResult (arg, actual, fpu, ulp, comment);
    return ulp;
    }

//----------------------------------------------------------------------------

int main (void)
    {
    sinTest (3.0162653335001840718L, "sin with smallest failing argument");
    sinTest (3.1415926535897932385L, "sin near pi");
    sinTest (4611686018694907549.0L, "sin with large argument");
    cosTest (1.5082562867317745453L, "cos with smallest failing argument");
    cosTest (1.5707963267948966193L, "cos near pi/2");
    cosTest (4611686018571982088.0L, "cos with large argument");
    tanTest (1.4430245999997931928L, "tan with smallest failing argument");
    tanTest (1.5707963267948966193L, "tan near pi/2");
    tanTest (4611686018694907549.0L, "tan with large argument");
    return 0;
    }

//----------------------------------------------------------------------------
